from typing import Optional

from .schedule import ScheduleOrFloat
from .transform import (
    add_decayed_weights,
    chain,
    identity,
    scale,
    scale_by_adam,
    scale_by_rms,
    scale_by_schedule,
    scale_by_stddev,
    trace,
)


def _scale_by_learning_rate(lr: ScheduleOrFloat):
    if callable(lr):
        return scale_by_schedule(lr)
    else:
        return scale(lr)


def sgd(learning_rate: ScheduleOrFloat = 1e-2, momentum: float = 0.9):
    return chain(
        trace(momentum),
        _scale_by_learning_rate(learning_rate),
    )


def rmsprop(
    learning_rate: ScheduleOrFloat = 1e-4,
    decay_rate: float = 0.9,
    eps: float = 1e-8,
    initial_scale: float = 0.0,
    centered: bool = False,
    momentum: Optional[float] = None,
):
    if centered:
        return chain(
            scale_by_stddev(
                decay_rate=decay_rate, eps=eps, initial_scale=initial_scale
            ),
            _scale_by_learning_rate(learning_rate),
            trace(decay_rate=momentum) if momentum is not None else identity(),
        )
    else:
        return chain(
            scale_by_rms(decay_rate=decay_rate, eps=eps, initial_scale=initial_scale),
            _scale_by_learning_rate(learning_rate),
            trace(decay_rate=momentum) if momentum is not None else identity(),
        )


def adam(
    learning_rate: ScheduleOrFloat = 1e-4,
    b1: float = 0.9,
    b2: float = 0.999,
    eps: float = 1e-8,
    eps_root: float = 0.0,
):
    return chain(
        scale_by_adam(b1=b1, b2=b2, eps=eps, eps_root=eps_root),
        _scale_by_learning_rate(learning_rate),
    )


def adamw(
    learning_rate: ScheduleOrFloat = 1e-4,
    b1: float = 0.9,
    b2: float = 0.999,
    eps: float = 1e-8,
    eps_root: float = 0.0,
    weight_decay: float = 1e-4,
):
    return chain(
        scale_by_adam(b1=b1, b2=b2, eps=eps, eps_root=eps_root),
        add_decayed_weights(weight_decay=weight_decay),
        _scale_by_learning_rate(learning_rate),
    )



import math
from typing import Callable, Optional, Union

import jax.numpy as jnp

ScheduleFn = Callable[[jnp.ndarray], jnp.ndarray]
ScheduleOrFloat = Union[Callable, float]


def exponential_decay(
    initial_value: float,
    step_size: float,
    reduce_factor: float = 2.0,
    min_value: Optional[float] = None,
) -> ScheduleOrFloat:
    """Exponential decay of lr.

    lr will be reduced by a factor of `reduce_factor` for every `step_size` steps.

    Arguments:
        initial_value: The initial value of learning rate.
        step_size: number of steps to reach a `reduce_factor` scaling.
        reduce_factor: the scaling factor per `step_size`.
        min_value: the minimum learning rate.
    """

    def _schedule_fn(step: jnp.ndarray):
        scale = 1.0 / step_size * math.log(reduce_factor)
        lr = jnp.exp(-step.astype(jnp.float32) * scale) * initial_value
        if min_value is not None:
            lr = jnp.clip(lr, a_min=min_value)
        return lr

    return _schedule_fn



"""Utility functions."""

from typing import Any, Optional, Tuple, TypeVar

import jax
import jax.numpy as jnp
import jmp
from pax import Module, assert_structure_equal, module_and_value

from .transform import GradientTransformation

TreeDef = Any

T = TypeVar("T")
K = TypeVar("K", bound=Module)
O = TypeVar("O", bound=Module)


def transform_gradients(grads: T, optimizer: O, params: T) -> Tuple[T, O]:
    """Transform gradients to updates using an optimizer.

    Arguments:
        grads: The gradients.
        optimizer: The gradient transformation.
        params: The trainable parameters.

    Returns:
        A pair ``(updates, optimizer)``

        - **updates** : The transformed gradients.
        - **optimizer** : The *updated* optimizer.
    """

    assert isinstance(optimizer, GradientTransformation), "Expecting an OPAX optimizer."
    optimizer, updates = module_and_value(optimizer)(grads, params=params)
    return updates, optimizer


def apply_updates(params: T, updates: T) -> T:
    """Update the parameters with updates.

    Arguments:
        params: The trainable parameters.
        updates: The transformed gradients.
    """
    assert_structure_equal(updates, params)
    return jax.tree_map(lambda u, p: p - u, updates, params)


def apply_gradients(
    model: T, optimizer: K, grads: T, all_finite: Optional[jnp.ndarray] = None
) -> Tuple[T, K]:
    """Update model and optimizer with gradients `grads`.

    Arguments:
        model: the model which contains trainable parameters.
        optimizer: the gradient transformation.
        grads: the gradients w.r.t to trainable parameters of `model`.
        all_finite: True if gradients are finite. Default: `None`.

    Returns:
        A pair ``(new_model, new_optimizer)``

        - **new_model**: the updated model.
        - **new_optimizer**: the updated optimizer.
    """
    assert isinstance(model, Module), "Expecting a PAX Module."
    assert type(model) is type(grads)
    params = model.parameters()
    updates, new_optimizer = transform_gradients(
        grads.parameters(), optimizer, params=params
    )
    new_params = apply_updates(params, updates=updates)

    if all_finite is not None:
        new_params, new_optimizer = jmp.select_tree(
            all_finite, (new_params, new_optimizer), (params, optimizer)
        )

    new_model = model.update_parameters(new_params)
    return new_model, new_optimizer



"""Gradient Transformations."""

from typing import Any, Callable, Sequence, TypeVar

import jax
import jax.numpy as jnp
import pax

T = TypeVar("T")


class GradientTransformation(pax.Module):
    def __init__(self, params=None):
        super().__init__()

    def __call__(self, updates, params=None):
        raise NotImplementedError("A subclass must implement this method")


def identity():
    class Identity(GradientTransformation):
        def __call__(self, updates, params=None):
            return updates

    return Identity


def scale(scale: float):
    class Scale(GradientTransformation):
        def __call__(self, updates, params=None):
            del params
            return jax.tree_map(lambda u: u * scale, updates)

    return Scale


def scale_by_schedule(schedule_fn: Callable[[jnp.ndarray], jnp.ndarray]):
    class ScaleBySchedule(GradientTransformation):
        count: jnp.ndarray
        learning_rate: jnp.ndarray

        def __init__(self, params):
            super().__init__(params=params)
            self.schedule_fn = schedule_fn
            self.register_state("count", jnp.array(0, dtype=jnp.int32))
            self.register_state("learning_rate", jnp.array(0.0, dtype=jnp.float32))

        def __call__(self, updates, params=None):
            del params
            self.count = self.count + 1
            self.learning_rate = self.schedule_fn(self.count)
            return jax.tree_map(lambda u: u * self.learning_rate, updates)

    return ScaleBySchedule


def scale_by_stddev(
    decay_rate: float = 0.9,
    eps: float = 1e-8,
    initial_scale: float = 0.0,
):
    class ScaleByStddev(GradientTransformation):
        def __init__(self, params):
            super().__init__(params=params)
            self.register_states(
                "mu",
                jax.tree_map(jnp.zeros_like, params),
            )
            self.register_states(
                "nu",
                jax.tree_map(lambda x: jnp.full_like(x, initial_scale), params),
            )

        def __call__(self, updates, params=None):
            del params
            self.mu = _update_moment(updates, self.mu, decay_rate, order=1)
            self.nu = _update_moment(updates, self.nu, decay_rate, order=2)

            updates = jax.tree_map(
                lambda g, m, n: g * jax.lax.rsqrt(n - jnp.square(m) + eps),
                updates,
                self.mu,
                self.nu,
            )

            return updates

    return ScaleByStddev


def scale_by_rms(
    decay_rate: float = 0.9,
    eps: float = 1e-8,
    initial_scale: float = 0.0,
):
    class ScaleByRms(GradientTransformation):
        def __init__(self, params):
            super().__init__(params=params)
            self.register_states(
                "nu", jax.tree_map(lambda x: jnp.full_like(x, initial_scale), params)
            )

        def __call__(self, updates, params=None):
            del params
            self.nu = _update_moment(updates, self.nu, decay_rate, order=2)
            updates = jax.tree_map(
                lambda g, n: g * jax.lax.rsqrt(n + eps),
                updates,
                self.nu,
            )
            return updates

    return ScaleByRms


def clip(max_delta: float):
    class Clip(GradientTransformation):
        def __call__(self, updates, params=None):
            del params
            return jax.tree_map(
                lambda u: jnp.clip(u, a_min=-max_delta, a_max=max_delta), updates
            )

    return Clip


def clip_by_global_norm(max_global_norm: float):
    class ClipByGlobalNorm(GradientTransformation):
        global_norm: jnp.ndarray  # for logging purposes.

        def __init__(self, params):
            super().__init__(params=params)
            self.register_state("global_norm", jnp.array(0.0))

        def __call__(self, updates, params=None):
            del params
            leaves = jax.tree_leaves(updates)
            leaves = jax.tree_map(lambda x: jnp.sum(jnp.square(x)), leaves)
            self.global_norm = jnp.sqrt(jnp.sum(jnp.stack(leaves)))
            scale = jnp.clip(max_global_norm / self.global_norm, a_max=1.0)
            return jax.tree_map(lambda x: x * scale, updates)

    return ClipByGlobalNorm


def trace(decay_rate):
    class Trace(GradientTransformation):
        trace: Any

        def __init__(self, params):
            super().__init__()

            self.register_states(
                "trace", jax.tree_map(lambda x: jnp.zeros_like(x), params)
            )

        def __call__(self, updates, params=None):
            self.trace = jax.tree_map(
                lambda u, t: u + t * decay_rate, updates, self.trace
            )
            return self.trace

    return Trace


def add_decayed_weights(weight_decay: float = 0.0):
    class AddDecayedWeights(GradientTransformation):
        def __init__(self, params):
            super().__init__(params=params)

        def __call__(self, updates, params=None):
            assert params is not None, "expecting params argument"

            updates = jax.tree_map(lambda g, p: g + weight_decay * p, updates, params)
            return updates

    return AddDecayedWeights


# source: https://github.com/deepmind/optax/blob/3f42a614096a7cd778e8cab15fd55e4766f47b53/optax/_src/transform.py#L78
def _update_moment(updates, moments, decay, order):
    """Compute the exponential moving average of the `order-th` moment."""
    return jax.tree_multimap(
        lambda g, t: (1 - decay) * (g ** order) + decay * t, updates, moments
    )


# source: https://github.com/deepmind/optax/blob/3f42a614096a7cd778e8cab15fd55e4766f47b53/optax/_src/transform.py#L84
def _bias_correction(moment, decay, count):
    """Perform bias correction. This becomes a no-op as count goes to infinity."""
    bias_correction = 1 - decay ** count
    return jax.tree_map(lambda t: t / bias_correction.astype(t.dtype), moment)


def ema(decay_rate: float, debias: bool = True):
    class EMA(GradientTransformation):
        ema: Any
        count: jnp.ndarray

        def __init__(self, params):
            super().__init__()
            self.register_state("count", jnp.array(0, dtype=jnp.int32))
            self.register_states("ema", jax.tree_map(jnp.zeros_like, params))

        def __call__(self, updates, params=None):
            del params
            self.ema = _update_moment(updates, self.ema, decay_rate, order=1)
            self.count = self.count + 1
            if debias:
                self.ema = _bias_correction(self.ema, decay_rate, self.count)
            return self.ema

    return EMA


def scale_by_adam(
    b1: float = 0.9,
    b2: float = 0.999,
    eps: float = 1e-8,
    eps_root: float = 0.0,
):
    class ScaleByAdam(GradientTransformation):
        mu: Any
        nu: Any
        count: jnp.ndarray

        def __init__(self, params):
            super().__init__(params)

            self.register_states("mu", jax.tree_map(jnp.zeros_like, params))
            self.register_states("nu", jax.tree_map(jnp.zeros_like, params))
            self.register_state("count", jnp.array(0, dtype=jnp.int32))

        def __call__(self, updates, params=None):
            del params
            self.mu = _update_moment(updates, self.mu, b1, order=1)
            self.nu = _update_moment(updates, self.nu, b2, order=2)
            self.count = self.count + 1
            mu_hat = _bias_correction(self.mu, b1, self.count)
            nu_hat = _bias_correction(self.nu, b2, self.count)

            updates = jax.tree_map(
                lambda m, v: m / (jnp.sqrt(v + eps_root) + eps), mu_hat, nu_hat
            )
            return updates

    return ScaleByAdam


def chain(*fs: Callable[[Any], GradientTransformation]):
    class Chain(GradientTransformation):
        transforms: Sequence[GradientTransformation]
        flatten: bool

        def __init__(self, params, flatten: bool = False):
            """Create a chain of gradient transformations.

            Arguments:
                params: trainable parameters.
                flatten: flatten trainable parameters to a list for faster speed in jit mode.
            """
            super().__init__()
            self.flatten = flatten
            if flatten:
                leaves = jax.tree_leaves(params)
                self.transforms = [f(leaves) for f in fs]
            else:
                self.transforms = [f(params) for f in fs]

        def __call__(self, updates, params=None):
            if self.flatten:
                updates_leaves, updates_treedef = jax.tree_flatten(updates)
                params_leaves = jax.tree_leaves(params)

                for f in self.transforms:
                    updates_leaves = f(updates=updates_leaves, params=params_leaves)

                updates = jax.tree_unflatten(updates_treedef, updates_leaves)
            else:
                for f in self.transforms:
                    updates = f(updates=updates, params=params)

            return updates

        def __getitem__(self, index: int):
            return self.transforms[index]

    return Chain



"""PAX optimizer library."""

from . import optimizer, schedule
from .optimizer import adam, adamw, rmsprop, sgd
from .transform import (
    GradientTransformation,
    chain,
    clip,
    clip_by_global_norm,
    scale,
    scale_by_adam,
    scale_by_rms,
    scale_by_schedule,
    scale_by_stddev,
    trace,
)
from .utils import apply_gradients, apply_updates, transform_gradients
